/************************************************************************/
/*                                                                      */
/* PROJEKT:    	Zugang		                                        */
/* PROGRAMM:   	root                                                   */
/*                                                                      */
/* AUTOR:      	Herrmann                                                */
/* VERSION:    	1.0 (20.01.92)                                          */
/*                                                                      */
/* FUNKTION:   							        */
/* BESCHREIBG: 								*/
/* 	laesst den user root-rechte bekommen				*/
/*                                                                      */
/* IN-PARAM:   			                                        */
/* Usage: d [-hsn] [-u user] [-c string] [name=value] ... [command [args] ]*/
/*                                                                      */
/* OUT-PARAM:                                                           */
/*									*/
/* AENDERUNG:  	Name   Datum   Umfang                                   */
/*                                                                      */
/*                                                                      */
/************************************************************************/


#define PATCHLEVEL	1
#include <stdio.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include "config.h"	/* configuration file */

#define DEBUG(level,msg)	if(debuglevel >= (level)) \
					(void) fprintf(stderr,"%s\n",msg);
#define error(string)	(void) fprintf(stderr,"%s: %s\n",progname,string);
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif
#define DEFAULTSHELL	"/bin/ksh"
#define ROOT_CONF	"/etc/root.conf"
#define USAGE	"Usage: d [-hsnV] [-u user] [-c string] [name=value] ... [command [args] ]\n"
char sccsid[]="%W% root.c (C) Copyright Herrmann";
char *PWD_ERR = "can't resolve password entry";

struct passwd *pwd;
char *shell = NULL;
char *progname = NULL;
int debuglevel = 0;
int dorootenv = TRUE;
int targetshell = FALSE;
int cdhome = FALSE;
int forward = FALSE;
char *username = NULL;

extern char *optarg;
extern int optind;

/*
 * Why aren't these objects declared
 * in standard header files???
 */
extern int putenv();
extern char *getenv();
extern char *getlogin();
extern struct passwd *getpwuid();
extern struct passwd *getpwnam();
extern char *strdup(); 
extern char **environ;

char *
basename(path)
char *path;
{
	char *cp;

	if(cp = strrchr(path,'/'))
		return cp+1;
	else	return path;
}

int
delenv(string)
char *string;
{
	char **env = environ;
	int len;

	for(;*env;env++) {
		len = strcspn(*env,"=");
		if(len != strlen(string))
			continue;
		if(strncmp(string,*env,len) == 0) {
			while(*env = *(env+1))
				env++;
			return 0;
		}
	}
	return -1;
}

main(argc,argv)
int argc;
char *argv[];
{
	char rootenv[1024];
	char line[256];
	char *home=NULL;	/* invoking (!) user's home directory */
	char *cp;
	char *name,*command;
	char logname[16];
	FILE *fp;
	int uid=0,gid=0;
	int c,endok;
	username="root";
	progname = basename(argv[0]);
	if(progname[0]=='+' && progname[1])
		username = progname + 1;

	if(geteuid() && getuid()) {
		(void) fprintf(stderr,"Sorry: %s ist nicht root und s!\n",progname);
		(void) exit(1);
	}

/*	while((c = getopt(argc-1,argv,"Vhsncxu?")) != EOF) { */
	while((c = getopt(argc,argv,"Vhsnc:x:u:?")) != EOF) {
		switch(c) {
		case 'c':	forward = TRUE;
				command = optarg;
				break;
		case 'h':	cdhome = TRUE;
				break;
		case 's':	targetshell = TRUE;
				break;
		case 'n':	dorootenv = FALSE;
				break;
		case 'u':	username = optarg;
				break;
		case 'x':	debuglevel = atoi(optarg);
				break;
		case 'V':	fprintf(stderr,"Version %d.%d Patchlevel %d\n",
					VERSION_MAYOR,VERSION_MINOR,PATCHLEVEL);
				exit(0);
				break;
		case '?':	
				fprintf(stderr,USAGE);
				exit(0);
				break;
		}
	}

	/*
	 * Determine login name by calling cuserid(3), getlogin(3) or
	 * if all that fails, getuid(2) and getpwuid(3).
	 */
	if(cuserid(logname) == EOF) {
		/*
		 * cuserid(3) failed, try getlogin(3)
		 */
		if(cp = getlogin()) {
			strcpy(logname,cp);
		} else {
			/*
			 * getlogin(3) failed too, try getpwuid(3)
			 */
			if((pwd = getpwuid(getuid())) == NULL) {
				(void) fprintf(stderr,
					"%s: kein Eintrag fuer Benutzer  %d\n",
					progname,getuid());
				error(PWD_ERR);
				(void) exit(1);
			}
			strcpy(logname,pwd->pw_name);
			home=strdup(pwd->pw_dir);
		}
	}
	/* nur dieser Aufruf ist sinnvoll */
	pwd = getpwuid(getuid());
	strcpy(logname,pwd->pw_name);
	home=strdup(pwd->pw_dir);

	if((pwd = getpwnam(username)) == NULL) {
		(void) fprintf(stderr,"%s: kein Eintrag fuer Benutzer %s\n",
			progname,username);
		error(PWD_ERR);
		(void) exit(1);
	}
	uid = pwd->pw_uid;
	gid = pwd->pw_gid;

	if(!CheckConfig(logname,username)) {
		(void) fprintf(stderr,"Sorry Benutzer %s darf das nicht\n",logname);
		(void) exit(1);
	}

	/*
	 * process $HOME/.{username}env file
	 *
	 * don't rely on $HOME, because it may be set to an
	 * different value already
	 */
	if(!home) {
		if((pwd = getpwnam(logname)) == NULL) {
			(void) fprintf(stderr,"%s: kein Eintrag fuer Benutzer %s\n",
				logname);
			exit(1);
		}
		home = strdup(pwd->pw_dir);
	}
	if(dorootenv && home) {
		(void) sprintf(rootenv,"%s/.%senv",home,username);
		if(fp = fopen(rootenv,"r")) {
			while(fgets(line,sizeof(line),fp)) {
				if(line[0] == '#')	/* comment */
					continue;
				if(cp = strrchr(line,'\n'))
					*cp = '\0';
				if(line[0] == '~' || line[0] == '^') {
					DEBUG(9,"removing environment string");
					DEBUG(9,line+1);
					(void) delenv(line+1);
				} else {
					cp = strdup(line);
					DEBUG(9,"adding environment string");
					DEBUG(9,cp);
					(void) putenv(cp);
				}
			}
			(void) fclose(fp);
		}
	}

	/*
	 * parse command line for environment strings
	 */

	while(optind < argc && strchr(argv[optind],'=')) {
		if(line[0] == '~' || line[0] == '^') {
			DEBUG(9,"removing environment string");
			DEBUG(9,line+1);
			(void) delenv(line+1);
		} else if(cp = strdup(argv[optind++])) {
			DEBUG(9,"adding environment string");
			DEBUG(9,cp);
			putenv(cp);
		}
	}

	if(forward || optind >= argc) {
		if(!targetshell)
			shell=getenv("SHELL");
		if(!shell) {
			if((pwd = getpwnam(targetshell ? username : logname)) == NULL) {
				error(PWD_ERR);
				(void) exit(1);
			}
			shell = pwd->pw_shell;
			if(*shell == '\0')
				shell = DEFAULTSHELL;
		}
		name = basename(shell);
	}

	/* Change identity */
	(void) setgid(gid);
	(void) setuid(uid);

	if(cdhome) {
		/*
		 * Change to NEW (!) home directory, therefore
		 * DO rely on $HOME!
		 */
		if(cp = getenv("HOME"))
			chdir(cp);
	}

	if(!forward && optind >= argc) {
		DEBUG(9,"executing shell");
		DEBUG(9,shell);
		DEBUG(9,name);
		(void) execl(shell,name,NULL);
	} else { /* commands specified */
		DEBUG(9,"executing command");
		if(forward) {
			endok= execl(shell,name,"-c",command,NULL);
		} else {
			endok= execvp(argv[optind],&argv[optind]);
		}
	}
	DEBUG(9,"exec failed");
	perror(progname);
	return 1;
    exit(endok);
}

CheckConfig(logname,username)
char *logname,*username;
{
	FILE *fp;
	int nfields;
	char *cp;
	char from[128],to[1024];

	/*
	 * The super user does not need an entry in ROOT_CONF
	 */
	if(getuid() == 0) 
		return TRUE;

	if((fp = fopen(ROOT_CONF,"r")) == NULL) {
#ifdef STRICT
		/* No configuration file found:
		 * This is a fatal error, no further actions
		 * may be taken.
		 */
		(void) fprintf(stderr,"%s: fatal error: can't open %s\n",
			progname,ROOT_CONF);
		return FALSE;
#else
		/* No configuration file found:
		 * everything is allowed.
		 */
		return TRUE;
#endif
	}

	while((nfields = fscanf(fp,"%s%[^\n]\n",from,to)) != EOF) { 
		if(from[0] == '#')
			continue;
		if(nfields != 2)
			continue;
#ifdef DEBUG1
		printf(" name: %s   to: %s  logname: %s username: %s\n",
		from,to,logname,username);
#endif
		if(strcmp(from,logname) == 0) {
			cp = strtok(to," ");
#ifdef DEBUG1
	 		printf(" cp: %s  \n",cp);
#endif
			do {
				DEBUG(5,cp);
				if(strcmp(cp,"*") == 0 ||
				strcmp(username,cp) == 0) {
					fclose(fp);
					return TRUE;
				}
			} while(cp = strtok(NULL,","));
		}
	}
			fclose(fp);
			return TRUE;
}
